(*
 * The Why certification tool
 * Copyright (C) 2002 Jean-Christophe FILLIATRE
 * 
 * This software is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public
 * License version 2, as published by the Free Software Foundation.
 * 
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * 
 * See the GNU General Public License version 2 for more details
 * (enclosed in the file GPL).
 *)

(*i $Id: parser.ml4,v 1.84 2003/06/18 15:11:02 filliatr Exp $ i*)

open Options
open Logic
open Rename
open Misc
open Util
open Types
open Ptree
open Env

(*s Lexer. Wrapper around [Plexer.make] to take an offset into account. *)

let offset = ref 0

let with_offset n f x =
  let old = !offset in
  try offset := n; let y = f x in offset := old; y with
    e -> offset := old; raise e

let loc_offset lf n = let (b, e) = lf n in b + !offset, e + !offset

let lexer =
  let l = Plexer.make () in
  {l with
    Token.func =
      fun cs -> let (ts, lf) = l.Token.func cs in ts, loc_offset lf}

(*s grammar entries *)

let gram = Grammar.create lexer
let gec s = Grammar.Entry.create gram s

(* logic *)
let lexpr = gec "lexpr"
let lexpr0 = gec "lexpr0"
let lexpr1 = gec "lexpr1"
let lexpr2 = gec "lexpr2"
let lexpr3 = gec "lexpr3"
let lexpr4 = gec "lexpr4"
let lexpr5 = gec "lexpr5"
let lexpr6 = gec "lexpr6"
let lexpr7 = gec "lexpr7"
let pp_relation = gec "pp_relation"
let constant = gec "constant"

(* types *)
let primitive_type = gec "primitive_type"
let type_v = gec "type_v"
let simple_type_v = gec "simple_type_v"
let type_c = gec "type_c"
let result = gec "result"
let effects = gec "effects"
let reads = gec "reads"
let writes = gec "writes"
let raises = gec "raises"
let cast = gec "cast"
let pre_condition = gec "pre_condition"
let post_condition = gec "post_condition"
let exn_condition = gec "exn_condition"

(* binders *)
let binder = gec "binder"
let binder_type = gec "binder_type"
let binders = gec "binders"

(* programs *)
let program = gec "program"
let prog1 = gec "prog1"
let prog2 = gec "prog2"
let prog3 = gec "prog3"
let prog4 = gec "prog4"
let plus_minus = gec "plus_minus"
let times_div_mod = gec "times_div_mod"
let prog5 = gec "prog5"
let prog6 = gec "prog6"
let prog7 = gec "prog7"
let recfun = gec "recfun"
let arg = gec "arg"
let block = gec "block"
let any_block = gec "any_block"
let block_statement = gec "block_statement"
let handler = gec "handler"
let relation = gec "relation"
let ident = gec "ident"
let qualid_ident = gec "qualid_ident"
let invariant = gec "invariant"
let variant = gec "variant"
let exception_type = gec "exception_type"
let assertion = gec "assertion"
let lexpr = gec "lexpr"
let name = gec "name"

let decl = gec "decl"
let decls = gec "decls"
let logic_type = gec "logic_type"
let logic_arg = gec "logic_arg"

(*s Utility functions. *)

let mk_pp loc d = {pp_loc = loc; pp_desc = d}
let infix_pp loc a i b = mk_pp loc (PPinfix (a, i, b))
let prefix_pp loc p a = mk_pp loc (PPprefix (p, a))

let conj_assert {a_name = n; a_value = a} {a_value = b} =
  let loc = Loc.join a.pp_loc b.pp_loc in
  {a_value = infix_pp loc a PPand b; a_name = n; a_loc = loc}

let conj =
  function
    None, None -> None
  | None, b -> b
  | a, None -> a
  | Some (a, []), Some (b, []) -> Some (conj_assert a b, [])
  | _ -> assert false (* TODO *)

let without_annot loc d =
  {pdesc = d; pre = []; oblig = []; post = None; ploc = loc}

let no_loc (_, e) = e, e

let rec app f =
  function
    [] -> assert false
  | [a] -> Sapp (f, a)
  | a :: l ->
      let loc = Loc.join f.ploc (arg_loc a) in
      app (without_annot loc (Sapp (f, a))) l

let bin_op (loc_op, op) loc e1 e2 =
  let f = without_annot loc_op (Svar op) in
  let f_e1 = without_annot (Loc.join e1.ploc loc_op) (Sapp (f, Sterm e1)) in
  without_annot loc (Sapp (f_e1, Sterm e2))

let un_op op loc e =
  without_annot loc (app (without_annot loc (Svar op)) [Sterm e])

let mk_prog loc p pre post =
  {pdesc = p.pdesc; pre = pre; oblig = p.pre @ p.oblig;
   post = conj (p.post, post); ploc = loc}

let rec_name =
  function
    Srec (x, _, _, _, _) -> x
  | _ -> assert false

let check_block loc b =
  let is_statement =
    function
      Sstatement _ -> true
    | _ -> false
  in
  if not (List.exists is_statement b) then
    Report.raise_located loc
      (Error.AnyMessage "a sequence must contain at least one statement")

let _ =
  Grammar.extend
    [Grammar.Entry.obj (ident : 'ident Grammar.Entry.e), None,
     [None, None,
      [[Gramext.Stoken ("UIDENT", "")],
       Gramext.action
         (fun (id : string) (loc : int * int) -> (Ident.create id : 'ident));
       [Gramext.Stoken ("LIDENT", "")],
       Gramext.action
         (fun (id : string) (loc : int * int) ->
            (Ident.create id : 'ident))]];
     Grammar.Entry.obj (qualid_ident : 'qualid_ident Grammar.Entry.e), None,
     [None, None,
      [[Gramext.Snterm (Grammar.Entry.obj (ident : 'ident Grammar.Entry.e));
        Gramext.Stoken ("", "@");
        Gramext.Snterm (Grammar.Entry.obj (ident : 'ident Grammar.Entry.e))],
       Gramext.action
         (fun (l : 'ident) _ (id : 'ident) (loc : int * int) ->
            (Ident.at_id id (Ident.string l) : 'qualid_ident));
       [Gramext.Snterm (Grammar.Entry.obj (ident : 'ident Grammar.Entry.e));
        Gramext.Stoken ("", "@")],
       Gramext.action
         (fun _ (id : 'ident) (loc : int * int) ->
            (Ident.at_id id "" : 'qualid_ident));
       [Gramext.Snterm (Grammar.Entry.obj (ident : 'ident Grammar.Entry.e))],
       Gramext.action
         (fun (id : 'ident) (loc : int * int) -> (id : 'qualid_ident))]];
     Grammar.Entry.obj (constant : 'constant Grammar.Entry.e), None,
     [None, None,
      [[Gramext.Stoken ("FLOAT", "")],
       Gramext.action
         (fun (f : string) (loc : int * int) -> (ConstFloat f : 'constant));
       [Gramext.Stoken ("LIDENT", "void")],
       Gramext.action (fun _ (loc : int * int) -> (ConstUnit : 'constant));
       [Gramext.Stoken ("", "false")],
       Gramext.action
         (fun _ (loc : int * int) -> (ConstBool false : 'constant));
       [Gramext.Stoken ("", "true")],
       Gramext.action
         (fun _ (loc : int * int) -> (ConstBool true : 'constant));
       [Gramext.Stoken ("INT", "")],
       Gramext.action
         (fun (n : string) (loc : int * int) ->
            (ConstInt (int_of_string n) : 'constant))]];
     Grammar.Entry.obj (lexpr : 'lexpr Grammar.Entry.e), None,
     [None, None,
      [[Gramext.Snterm
          (Grammar.Entry.obj (lexpr0 : 'lexpr0 Grammar.Entry.e))],
       Gramext.action (fun (a : 'lexpr0) (loc : int * int) -> (a : 'lexpr));
       [Gramext.Snterm (Grammar.Entry.obj (lexpr0 : 'lexpr0 Grammar.Entry.e));
        Gramext.Stoken ("", "->"); Gramext.Sself],
       Gramext.action
         (fun (b : 'lexpr) _ (a : 'lexpr0) (loc : int * int) ->
            (infix_pp loc a PPimplies b : 'lexpr))]];
     Grammar.Entry.obj (lexpr0 : 'lexpr0 Grammar.Entry.e), None,
     [None, None,
      [[Gramext.Snterm
          (Grammar.Entry.obj (lexpr1 : 'lexpr1 Grammar.Entry.e))],
       Gramext.action (fun (a : 'lexpr1) (loc : int * int) -> (a : 'lexpr0));
       [Gramext.Snterm (Grammar.Entry.obj (lexpr1 : 'lexpr1 Grammar.Entry.e));
        Gramext.Stoken ("", "or"); Gramext.Sself],
       Gramext.action
         (fun (b : 'lexpr0) _ (a : 'lexpr1) (loc : int * int) ->
            (infix_pp loc a PPor b : 'lexpr0))]];
     Grammar.Entry.obj (lexpr1 : 'lexpr1 Grammar.Entry.e), None,
     [None, None,
      [[Gramext.Snterm
          (Grammar.Entry.obj (lexpr2 : 'lexpr2 Grammar.Entry.e))],
       Gramext.action (fun (a : 'lexpr2) (loc : int * int) -> (a : 'lexpr1));
       [Gramext.Snterm (Grammar.Entry.obj (lexpr2 : 'lexpr2 Grammar.Entry.e));
        Gramext.Stoken ("", "and"); Gramext.Sself],
       Gramext.action
         (fun (b : 'lexpr1) _ (a : 'lexpr2) (loc : int * int) ->
            (infix_pp loc a PPand b : 'lexpr1))]];
     Grammar.Entry.obj (lexpr2 : 'lexpr2 Grammar.Entry.e), None,
     [None, None,
      [[Gramext.Snterm
          (Grammar.Entry.obj (lexpr3 : 'lexpr3 Grammar.Entry.e))],
       Gramext.action (fun (a : 'lexpr3) (loc : int * int) -> (a : 'lexpr2));
       [Gramext.Stoken ("LIDENT", "not");
        Gramext.Snterm
          (Grammar.Entry.obj (lexpr3 : 'lexpr3 Grammar.Entry.e))],
       Gramext.action
         (fun (a : 'lexpr3) _ (loc : int * int) ->
            (prefix_pp loc PPnot a : 'lexpr2))]];
     Grammar.Entry.obj (lexpr3 : 'lexpr3 Grammar.Entry.e), None,
     [None, None,
      [[Gramext.Snterm
          (Grammar.Entry.obj (lexpr4 : 'lexpr4 Grammar.Entry.e))],
       Gramext.action (fun (a : 'lexpr4) (loc : int * int) -> (a : 'lexpr3));
       [Gramext.Snterm (Grammar.Entry.obj (lexpr4 : 'lexpr4 Grammar.Entry.e));
        Gramext.Snterm
          (Grammar.Entry.obj (pp_relation : 'pp_relation Grammar.Entry.e));
        Gramext.Snterm (Grammar.Entry.obj (lexpr4 : 'lexpr4 Grammar.Entry.e));
        Gramext.Snterm
          (Grammar.Entry.obj (pp_relation : 'pp_relation Grammar.Entry.e));
        Gramext.Snterm
          (Grammar.Entry.obj (lexpr4 : 'lexpr4 Grammar.Entry.e))],
       Gramext.action
         (fun (c : 'lexpr4) (r2 : 'pp_relation) (b : 'lexpr4)
            (r1 : 'pp_relation) (a : 'lexpr4) (loc : int * int) ->
            (infix_pp loc (infix_pp loc a r1 b) PPand (infix_pp loc b r2 c) :
             'lexpr3));
       [Gramext.Snterm (Grammar.Entry.obj (lexpr4 : 'lexpr4 Grammar.Entry.e));
        Gramext.Snterm
          (Grammar.Entry.obj (pp_relation : 'pp_relation Grammar.Entry.e));
        Gramext.Snterm
          (Grammar.Entry.obj (lexpr4 : 'lexpr4 Grammar.Entry.e))],
       Gramext.action
         (fun (b : 'lexpr4) (r : 'pp_relation) (a : 'lexpr4)
            (loc : int * int) ->
            (infix_pp loc a r b : 'lexpr3))]];
     Grammar.Entry.obj (lexpr4 : 'lexpr4 Grammar.Entry.e), None,
     [None, None,
      [[Gramext.Snterm
          (Grammar.Entry.obj (lexpr5 : 'lexpr5 Grammar.Entry.e))],
       Gramext.action (fun (a : 'lexpr5) (loc : int * int) -> (a : 'lexpr4));
       [Gramext.Sself; Gramext.Stoken ("", "-");
        Gramext.Snterm
          (Grammar.Entry.obj (lexpr5 : 'lexpr5 Grammar.Entry.e))],
       Gramext.action
         (fun (b : 'lexpr5) _ (a : 'lexpr4) (loc : int * int) ->
            (infix_pp loc a PPsub b : 'lexpr4));
       [Gramext.Sself; Gramext.Stoken ("", "+");
        Gramext.Snterm
          (Grammar.Entry.obj (lexpr5 : 'lexpr5 Grammar.Entry.e))],
       Gramext.action
         (fun (b : 'lexpr5) _ (a : 'lexpr4) (loc : int * int) ->
            (infix_pp loc a PPadd b : 'lexpr4))]];
     Grammar.Entry.obj (lexpr5 : 'lexpr5 Grammar.Entry.e), None,
     [None, None,
      [[Gramext.Snterm
          (Grammar.Entry.obj (lexpr6 : 'lexpr6 Grammar.Entry.e))],
       Gramext.action (fun (a : 'lexpr6) (loc : int * int) -> (a : 'lexpr5));
       [Gramext.Sself; Gramext.Stoken ("", "%");
        Gramext.Snterm
          (Grammar.Entry.obj (lexpr6 : 'lexpr6 Grammar.Entry.e))],
       Gramext.action
         (fun (b : 'lexpr6) _ (a : 'lexpr5) (loc : int * int) ->
            (infix_pp loc a PPmod b : 'lexpr5));
       [Gramext.Sself; Gramext.Stoken ("", "/");
        Gramext.Snterm
          (Grammar.Entry.obj (lexpr6 : 'lexpr6 Grammar.Entry.e))],
       Gramext.action
         (fun (b : 'lexpr6) _ (a : 'lexpr5) (loc : int * int) ->
            (infix_pp loc a PPdiv b : 'lexpr5));
       [Gramext.Sself; Gramext.Stoken ("", "*");
        Gramext.Snterm
          (Grammar.Entry.obj (lexpr6 : 'lexpr6 Grammar.Entry.e))],
       Gramext.action
         (fun (b : 'lexpr6) _ (a : 'lexpr5) (loc : int * int) ->
            (infix_pp loc a PPmul b : 'lexpr5))]];
     Grammar.Entry.obj (lexpr6 : 'lexpr6 Grammar.Entry.e), None,
     [None, None,
      [[Gramext.Snterm
          (Grammar.Entry.obj (lexpr7 : 'lexpr7 Grammar.Entry.e))],
       Gramext.action (fun (a : 'lexpr7) (loc : int * int) -> (a : 'lexpr6));
       [Gramext.Stoken ("", "-"); Gramext.Sself],
       Gramext.action
         (fun (a : 'lexpr6) _ (loc : int * int) ->
            (prefix_pp loc PPneg a : 'lexpr6))]];
     Grammar.Entry.obj (lexpr7 : 'lexpr7 Grammar.Entry.e), None,
     [None, None,
      [[Gramext.Stoken ("", "(");
        Gramext.Snterm (Grammar.Entry.obj (lexpr : 'lexpr Grammar.Entry.e));
        Gramext.Stoken ("", ")")],
       Gramext.action
         (fun _ (a : 'lexpr) _ (loc : int * int) -> (a : 'lexpr7));
       [Gramext.Stoken ("LIDENT", "exists");
        Gramext.Snterm (Grammar.Entry.obj (ident : 'ident Grammar.Entry.e));
        Gramext.Stoken ("", ":");
        Gramext.Snterm
          (Grammar.Entry.obj
             (primitive_type : 'primitive_type Grammar.Entry.e));
        Gramext.Stoken ("", ".");
        Gramext.Snterm (Grammar.Entry.obj (lexpr : 'lexpr Grammar.Entry.e))],
       Gramext.action
         (fun (a : 'lexpr) _ (t : 'primitive_type) _ (id : 'ident) _
            (loc : int * int) ->
            (mk_pp loc (PPexists (id, t, a)) : 'lexpr7));
       [Gramext.Stoken ("LIDENT", "forall");
        Gramext.Snterm (Grammar.Entry.obj (ident : 'ident Grammar.Entry.e));
        Gramext.Stoken ("", ":");
        Gramext.Snterm
          (Grammar.Entry.obj
             (primitive_type : 'primitive_type Grammar.Entry.e));
        Gramext.Stoken ("", ".");
        Gramext.Snterm (Grammar.Entry.obj (lexpr : 'lexpr Grammar.Entry.e))],
       Gramext.action
         (fun (a : 'lexpr) _ (t : 'primitive_type) _ (id : 'ident) _
            (loc : int * int) ->
            (mk_pp loc (PPforall (id, t, a)) : 'lexpr7));
       [Gramext.Stoken ("", "if");
        Gramext.Snterm (Grammar.Entry.obj (lexpr : 'lexpr Grammar.Entry.e));
        Gramext.Stoken ("", "then");
        Gramext.Snterm (Grammar.Entry.obj (lexpr : 'lexpr Grammar.Entry.e));
        Gramext.Stoken ("", "else");
        Gramext.Snterm (Grammar.Entry.obj (lexpr : 'lexpr Grammar.Entry.e))],
       Gramext.action
         (fun (p2 : 'lexpr) _ (p1 : 'lexpr) _ (p0 : 'lexpr) _
            (loc : int * int) ->
            (mk_pp loc (PPif (p0, p1, p2)) : 'lexpr7));
       [Gramext.Snterm
          (Grammar.Entry.obj (qualid_ident : 'qualid_ident Grammar.Entry.e));
        Gramext.Stoken ("", "[");
        Gramext.Snterm (Grammar.Entry.obj (lexpr : 'lexpr Grammar.Entry.e));
        Gramext.Stoken ("", "]")],
       Gramext.action
         (fun _ (t : 'lexpr) _ (x : 'qualid_ident) (loc : int * int) ->
            (mk_pp loc (PPapp (Ident.access, [mk_pp loc (PPvar x); t])) :
             'lexpr7));
       [Gramext.Snterm
          (Grammar.Entry.obj (qualid_ident : 'qualid_ident Grammar.Entry.e));
        Gramext.Stoken ("", "(");
        Gramext.Slist1sep
          (Gramext.Snterm
             (Grammar.Entry.obj (lexpr : 'lexpr Grammar.Entry.e)),
           Gramext.Stoken ("", ","));
        Gramext.Stoken ("", ")")],
       Gramext.action
         (fun _ (l : 'lexpr list) _ (x : 'qualid_ident) (loc : int * int) ->
            (mk_pp loc (PPapp (x, l)) : 'lexpr7));
       [Gramext.Snterm
          (Grammar.Entry.obj (qualid_ident : 'qualid_ident Grammar.Entry.e))],
       Gramext.action
         (fun (x : 'qualid_ident) (loc : int * int) ->
            (mk_pp loc (PPvar x) : 'lexpr7));
       [Gramext.Snterm
          (Grammar.Entry.obj (constant : 'constant Grammar.Entry.e))],
       Gramext.action
         (fun (c : 'constant) (loc : int * int) ->
            (mk_pp loc (PPconst c) : 'lexpr7));
       [Gramext.Stoken ("", "false")],
       Gramext.action
         (fun _ (loc : int * int) -> (mk_pp loc PPfalse : 'lexpr7));
       [Gramext.Stoken ("", "true")],
       Gramext.action
         (fun _ (loc : int * int) -> (mk_pp loc PPtrue : 'lexpr7))]];
     Grammar.Entry.obj (pp_relation : 'pp_relation Grammar.Entry.e), None,
     [None, None,
      [[Gramext.Stoken ("", "<>")],
       Gramext.action (fun _ (loc : int * int) -> (PPneq : 'pp_relation));
       [Gramext.Stoken ("", "=")],
       Gramext.action (fun _ (loc : int * int) -> (PPeq : 'pp_relation));
       [Gramext.Stoken ("", ">=")],
       Gramext.action (fun _ (loc : int * int) -> (PPge : 'pp_relation));
       [Gramext.Stoken ("", ">")],
       Gramext.action (fun _ (loc : int * int) -> (PPgt : 'pp_relation));
       [Gramext.Stoken ("", "<=")],
       Gramext.action (fun _ (loc : int * int) -> (PPle : 'pp_relation));
       [Gramext.Stoken ("", "<")],
       Gramext.action (fun _ (loc : int * int) -> (PPlt : 'pp_relation))]];
     Grammar.Entry.obj (primitive_type : 'primitive_type Grammar.Entry.e),
     None,
     [None, None,
      [[Gramext.Snterm (Grammar.Entry.obj (ident : 'ident Grammar.Entry.e))],
       Gramext.action
         (fun (id : 'ident) (loc : int * int) ->
            (PTexternal id : 'primitive_type));
       [Gramext.Stoken ("", "unit")],
       Gramext.action (fun _ (loc : int * int) -> (PTunit : 'primitive_type));
       [Gramext.Stoken ("", "float")],
       Gramext.action
         (fun _ (loc : int * int) -> (PTfloat : 'primitive_type));
       [Gramext.Stoken ("", "bool")],
       Gramext.action (fun _ (loc : int * int) -> (PTbool : 'primitive_type));
       [Gramext.Stoken ("", "int")],
       Gramext.action
         (fun _ (loc : int * int) -> (PTint : 'primitive_type))]];
     Grammar.Entry.obj (type_v : 'type_v Grammar.Entry.e), None,
     [None, None,
      [[Gramext.Snterm
          (Grammar.Entry.obj
             (simple_type_v : 'simple_type_v Grammar.Entry.e))],
       Gramext.action
         (fun (t : 'simple_type_v) (loc : int * int) -> (t : 'type_v));
       [Gramext.Stoken ("UIDENT", ""); Gramext.Stoken ("", ":");
        Gramext.Snterm
          (Grammar.Entry.obj
             (simple_type_v : 'simple_type_v Grammar.Entry.e));
        Gramext.Stoken ("", "->");
        Gramext.Snterm
          (Grammar.Entry.obj (type_c : 'type_c Grammar.Entry.e))],
       Gramext.action
         (fun (c : 'type_c) _ (v : 'simple_type_v) _ (x : string)
            (loc : int * int) ->
            (PVarrow ([Ident.create x, BindType v], c) : 'type_v));
       [Gramext.Stoken ("LIDENT", ""); Gramext.Stoken ("", ":");
        Gramext.Snterm
          (Grammar.Entry.obj
             (simple_type_v : 'simple_type_v Grammar.Entry.e));
        Gramext.Stoken ("", "->");
        Gramext.Snterm
          (Grammar.Entry.obj (type_c : 'type_c Grammar.Entry.e))],
       Gramext.action
         (fun (c : 'type_c) _ (v : 'simple_type_v) _ (x : string)
            (loc : int * int) ->
            (PVarrow ([Ident.create x, BindType v], c) : 'type_v));
       [Gramext.Snterm
          (Grammar.Entry.obj
             (simple_type_v : 'simple_type_v Grammar.Entry.e));
        Gramext.Stoken ("", "->");
        Gramext.Snterm
          (Grammar.Entry.obj (type_c : 'type_c Grammar.Entry.e))],
       Gramext.action
         (fun (c : 'type_c) _ (v : 'simple_type_v) (loc : int * int) ->
            (PVarrow ([Ident.anonymous, BindType v], c) : 'type_v))]];
     Grammar.Entry.obj (simple_type_v : 'simple_type_v Grammar.Entry.e), None,
     [None, None,
      [[Gramext.Stoken ("", "(");
        Gramext.Snterm (Grammar.Entry.obj (type_v : 'type_v Grammar.Entry.e));
        Gramext.Stoken ("", ")")],
       Gramext.action
         (fun _ (v : 'type_v) _ (loc : int * int) -> (v : 'simple_type_v));
       [Gramext.Snterm
          (Grammar.Entry.obj
             (primitive_type : 'primitive_type Grammar.Entry.e))],
       Gramext.action
         (fun (t : 'primitive_type) (loc : int * int) ->
            (PVpure t : 'simple_type_v));
       [Gramext.Sself; Gramext.Stoken ("", "ref")],
       Gramext.action
         (fun _ (v : 'simple_type_v) (loc : int * int) ->
            (PVref v : 'simple_type_v));
       [Gramext.Sself; Gramext.Stoken ("", "array")],
       Gramext.action
         (fun _ (v : 'simple_type_v) (loc : int * int) ->
            (PVarray v : 'simple_type_v))]];
     Grammar.Entry.obj (type_c : 'type_c Grammar.Entry.e), None,
     [None, None,
      [[Gramext.Snterm
          (Grammar.Entry.obj (type_v : 'type_v Grammar.Entry.e))],
       Gramext.action
         (fun (v : 'type_v) (loc : int * int) -> (ptype_c_of_v v : 'type_c));
       [Gramext.Stoken ("", "{");
        Gramext.Sopt
          (Gramext.Snterm
             (Grammar.Entry.obj
                (pre_condition : 'pre_condition Grammar.Entry.e)));
        Gramext.Stoken ("", "}");
        Gramext.Snterm (Grammar.Entry.obj (result : 'result Grammar.Entry.e));
        Gramext.Snterm
          (Grammar.Entry.obj (effects : 'effects Grammar.Entry.e));
        Gramext.Stoken ("", "{");
        Gramext.Sopt
          (Gramext.Snterm
             (Grammar.Entry.obj
                (post_condition : 'post_condition Grammar.Entry.e)));
        Gramext.Stoken ("", "}")],
       Gramext.action
         (fun _ (q : 'post_condition option) _ (e : 'effects)
            (id, v : 'result) _ (p : 'pre_condition option) _
            (loc : int * int) ->
            ({pc_result_name = id; pc_result_type = v; pc_effect = e;
              pc_pre = list_of_some p; pc_post = q} :
             'type_c))]];
     Grammar.Entry.obj (result : 'result Grammar.Entry.e), None,
     [None, None,
      [[Gramext.Snterm
          (Grammar.Entry.obj (type_v : 'type_v Grammar.Entry.e))],
       Gramext.action
         (fun (v : 'type_v) (loc : int * int) -> (Ident.result, v : 'result));
       [Gramext.Stoken ("LIDENT", "returns");
        Gramext.Snterm (Grammar.Entry.obj (ident : 'ident Grammar.Entry.e));
        Gramext.Stoken ("", ":");
        Gramext.Snterm
          (Grammar.Entry.obj (type_v : 'type_v Grammar.Entry.e))],
       Gramext.action
         (fun (v : 'type_v) _ (id : 'ident) _ (loc : int * int) ->
            (id, v : 'result))]];
     Grammar.Entry.obj (effects : 'effects Grammar.Entry.e), None,
     [None, None,
      [[Gramext.Sopt
          (Gramext.Snterm
             (Grammar.Entry.obj (reads : 'reads Grammar.Entry.e)));
        Gramext.Sopt
          (Gramext.Snterm
             (Grammar.Entry.obj (writes : 'writes Grammar.Entry.e)));
        Gramext.Sopt
          (Gramext.Snterm
             (Grammar.Entry.obj (raises : 'raises Grammar.Entry.e)))],
       Gramext.action
         (fun (x : 'raises option) (w : 'writes option) (r : 'reads option)
            (loc : int * int) ->
            (let r' =
               match r with
                 Some l -> l
               | _ -> []
             in
             let w' =
               match w with
                 Some l -> l
               | _ -> []
             in
             let x' =
               match x with
                 Some l -> l
               | _ -> []
             in
             List.fold_right Effect.add_write w'
               (List.fold_right Effect.add_read r'
                  (List.fold_right Effect.add_exn x' Effect.bottom)) :
             'effects))]];
     Grammar.Entry.obj (reads : 'reads Grammar.Entry.e), None,
     [None, None,
      [[Gramext.Stoken ("LIDENT", "reads");
        Gramext.Slist0sep
          (Gramext.Snterm
             (Grammar.Entry.obj (ident : 'ident Grammar.Entry.e)),
           Gramext.Stoken ("", ","))],
       Gramext.action
         (fun (l : 'ident list) _ (loc : int * int) -> (l : 'reads))]];
     Grammar.Entry.obj (writes : 'writes Grammar.Entry.e), None,
     [None, None,
      [[Gramext.Stoken ("LIDENT", "writes");
        Gramext.Slist0sep
          (Gramext.Snterm
             (Grammar.Entry.obj (ident : 'ident Grammar.Entry.e)),
           Gramext.Stoken ("", ","))],
       Gramext.action
         (fun (l : 'ident list) _ (loc : int * int) -> (l : 'writes))]];
     Grammar.Entry.obj (raises : 'raises Grammar.Entry.e), None,
     [None, None,
      [[Gramext.Stoken ("LIDENT", "raises");
        Gramext.Slist0sep
          (Gramext.Snterm
             (Grammar.Entry.obj (ident : 'ident Grammar.Entry.e)),
           Gramext.Stoken ("", ","))],
       Gramext.action
         (fun (l : 'ident list) _ (loc : int * int) -> (l : 'raises))]];
     Grammar.Entry.obj (pre_condition : 'pre_condition Grammar.Entry.e), None,
     [None, None,
      [[Gramext.Snterm
          (Grammar.Entry.obj (assertion : 'assertion Grammar.Entry.e))],
       Gramext.action
         (fun (c : 'assertion) (loc : int * int) -> (c : 'pre_condition))]];
     Grammar.Entry.obj (post_condition : 'post_condition Grammar.Entry.e),
     None,
     [None, None,
      [[Gramext.Stoken ("", "|");
        Gramext.Slist1sep
          (Gramext.Snterm
             (Grammar.Entry.obj
                (exn_condition : 'exn_condition Grammar.Entry.e)),
           Gramext.Stoken ("", "|"))],
       Gramext.action
         (fun (l : 'exn_condition list) _ (loc : int * int) ->
            (wprintf loc "no postcondition; false inserted@\n";
             if werror then exit 1;
             anonymous loc (mk_pp loc PPfalse), l :
             'post_condition));
       [Gramext.Snterm
          (Grammar.Entry.obj (assertion : 'assertion Grammar.Entry.e));
        Gramext.Stoken ("", "|");
        Gramext.Slist1sep
          (Gramext.Snterm
             (Grammar.Entry.obj
                (exn_condition : 'exn_condition Grammar.Entry.e)),
           Gramext.Stoken ("", "|"))],
       Gramext.action
         (fun (l : 'exn_condition list) _ (c : 'assertion)
            (loc : int * int) ->
            (c, l : 'post_condition));
       [Gramext.Snterm
          (Grammar.Entry.obj (assertion : 'assertion Grammar.Entry.e))],
       Gramext.action
         (fun (c : 'assertion) (loc : int * int) ->
            (c, [] : 'post_condition))]];
     Grammar.Entry.obj (exn_condition : 'exn_condition Grammar.Entry.e), None,
     [None, None,
      [[Gramext.Snterm (Grammar.Entry.obj (ident : 'ident Grammar.Entry.e));
        Gramext.Stoken ("", "=>");
        Gramext.Snterm
          (Grammar.Entry.obj (assertion : 'assertion Grammar.Entry.e))],
       Gramext.action
         (fun (c : 'assertion) _ (x : 'ident) (loc : int * int) ->
            (x, c : 'exn_condition))]];
     Grammar.Entry.obj (binder : 'binder Grammar.Entry.e), None,
     [None, None,
      [[Gramext.Stoken ("", "(");
        Gramext.Slist1sep
          (Gramext.Snterm
             (Grammar.Entry.obj (ident : 'ident Grammar.Entry.e)),
           Gramext.Stoken ("", ","));
        Gramext.Stoken ("", ":");
        Gramext.Snterm
          (Grammar.Entry.obj (binder_type : 'binder_type Grammar.Entry.e));
        Gramext.Stoken ("", ")")],
       Gramext.action
         (fun _ (t : 'binder_type) _ (sl : 'ident list) _ (loc : int * int) ->
            (List.map (fun s -> s, t) sl : 'binder))]];
     Grammar.Entry.obj (binder_type : 'binder_type Grammar.Entry.e), None,
     [None, None,
      [[Gramext.Snterm
          (Grammar.Entry.obj (type_v : 'type_v Grammar.Entry.e))],
       Gramext.action
         (fun (v : 'type_v) (loc : int * int) -> (BindType v : 'binder_type));
       [Gramext.Stoken ("UIDENT", "Set")],
       Gramext.action (fun _ (loc : int * int) -> (BindSet : 'binder_type))]];
     Grammar.Entry.obj (binders : 'binders Grammar.Entry.e), None,
     [None, None,
      [[Gramext.Slist1
          (Gramext.Snterm
             (Grammar.Entry.obj (binder : 'binder Grammar.Entry.e)))],
       Gramext.action
         (fun (bl : 'binder list) (loc : int * int) ->
            (List.flatten bl : 'binders))]];
     Grammar.Entry.obj (assertion : 'assertion Grammar.Entry.e), None,
     [None, None,
      [[Gramext.Snterm (Grammar.Entry.obj (lexpr : 'lexpr Grammar.Entry.e));
        Gramext.Snterm (Grammar.Entry.obj (name : 'name Grammar.Entry.e))],
       Gramext.action
         (fun (n : 'name) (c : 'lexpr) (loc : int * int) ->
            ({a_name = n; a_value = c; a_loc = loc} : 'assertion))]];
     Grammar.Entry.obj (name : 'name Grammar.Entry.e), None,
     [None, None,
      [[],
       Gramext.action (fun (loc : int * int) -> (Ident.Anonymous : 'name));
       [Gramext.Stoken ("", "as");
        Gramext.Snterm (Grammar.Entry.obj (ident : 'ident Grammar.Entry.e))],
       Gramext.action
         (fun (id : 'ident) _ (loc : int * int) -> (Ident.Name id : 'name))]];
     Grammar.Entry.obj (program : 'program Grammar.Entry.e), None,
     [None, None,
      [[Gramext.Snterm (Grammar.Entry.obj (prog1 : 'prog1 Grammar.Entry.e))],
       Gramext.action
         (fun (p : 'prog1) (loc : int * int) -> (p : 'program))]];
     Grammar.Entry.obj (prog1 : 'prog1 Grammar.Entry.e), None,
     [None, None,
      [[Gramext.Snterm (Grammar.Entry.obj (prog2 : 'prog2 Grammar.Entry.e))],
       Gramext.action (fun (x : 'prog2) (loc : int * int) -> (x : 'prog1));
       [Gramext.Snterm (Grammar.Entry.obj (prog2 : 'prog2 Grammar.Entry.e));
        Gramext.Stoken ("", "&&"); Gramext.Sself],
       Gramext.action
         (fun (y : 'prog1) _ (x : 'prog2) (loc : int * int) ->
            (let pf = without_annot (no_loc loc) (Sconst (ConstBool false)) in
             without_annot loc (Sif (x, y, pf)) :
             'prog1));
       [Gramext.Snterm (Grammar.Entry.obj (prog2 : 'prog2 Grammar.Entry.e));
        Gramext.Stoken ("", "||"); Gramext.Sself],
       Gramext.action
         (fun (y : 'prog1) _ (x : 'prog2) (loc : int * int) ->
            (let ptrue =
               without_annot (no_loc loc) (Sconst (ConstBool true))
             in
             without_annot loc (Sif (x, ptrue, y)) :
             'prog1))]];
     Grammar.Entry.obj (prog2 : 'prog2 Grammar.Entry.e), None,
     [None, None,
      [[Gramext.Snterm (Grammar.Entry.obj (prog3 : 'prog3 Grammar.Entry.e))],
       Gramext.action (fun (x : 'prog3) (loc : int * int) -> (x : 'prog2));
       [Gramext.Stoken ("LIDENT", "not");
        Gramext.Snterm (Grammar.Entry.obj (prog3 : 'prog3 Grammar.Entry.e))],
       Gramext.action
         (fun (x : 'prog3) _ (loc : int * int) ->
            (let pf = without_annot (no_loc loc) (Sconst (ConstBool false)) in
             let pt = without_annot (no_loc loc) (Sconst (ConstBool true)) in
             without_annot loc (Sif (x, pf, pt)) :
             'prog2))]];
     Grammar.Entry.obj (prog3 : 'prog3 Grammar.Entry.e), None,
     [None, None,
      [[Gramext.Snterm (Grammar.Entry.obj (prog4 : 'prog4 Grammar.Entry.e))],
       Gramext.action (fun (x : 'prog4) (loc : int * int) -> (x : 'prog3));
       [Gramext.Snterm (Grammar.Entry.obj (prog4 : 'prog4 Grammar.Entry.e));
        Gramext.Snterm
          (Grammar.Entry.obj (relation : 'relation Grammar.Entry.e));
        Gramext.Snterm (Grammar.Entry.obj (prog4 : 'prog4 Grammar.Entry.e))],
       Gramext.action
         (fun (y : 'prog4) (rel : 'relation) (x : 'prog4) (loc : int * int) ->
            (bin_op rel loc x y : 'prog3))]];
     Grammar.Entry.obj (prog4 : 'prog4 Grammar.Entry.e), None,
     [None, None,
      [[Gramext.Snterm (Grammar.Entry.obj (prog5 : 'prog5 Grammar.Entry.e))],
       Gramext.action (fun (x : 'prog5) (loc : int * int) -> (x : 'prog4));
       [Gramext.Snterm (Grammar.Entry.obj (prog5 : 'prog5 Grammar.Entry.e));
        Gramext.Snterm
          (Grammar.Entry.obj (plus_minus : 'plus_minus Grammar.Entry.e));
        Gramext.Sself],
       Gramext.action
         (fun (y : 'prog4) (op : 'plus_minus) (x : 'prog5)
            (loc : int * int) ->
            (bin_op op loc x y : 'prog4))]];
     Grammar.Entry.obj (plus_minus : 'plus_minus Grammar.Entry.e), None,
     [None, None,
      [[Gramext.Stoken ("", "-")],
       Gramext.action
         (fun _ (loc : int * int) -> (loc, Ident.t_sub : 'plus_minus));
       [Gramext.Stoken ("", "+")],
       Gramext.action
         (fun _ (loc : int * int) -> (loc, Ident.t_add : 'plus_minus))]];
     Grammar.Entry.obj (prog5 : 'prog5 Grammar.Entry.e), None,
     [None, None,
      [[Gramext.Snterm (Grammar.Entry.obj (prog6 : 'prog6 Grammar.Entry.e))],
       Gramext.action (fun (x : 'prog6) (loc : int * int) -> (x : 'prog5));
       [Gramext.Snterm (Grammar.Entry.obj (prog6 : 'prog6 Grammar.Entry.e));
        Gramext.Snterm
          (Grammar.Entry.obj
             (times_div_mod : 'times_div_mod Grammar.Entry.e));
        Gramext.Sself],
       Gramext.action
         (fun (y : 'prog5) (op : 'times_div_mod) (x : 'prog6)
            (loc : int * int) ->
            (bin_op op loc x y : 'prog5))]];
     Grammar.Entry.obj (times_div_mod : 'times_div_mod Grammar.Entry.e), None,
     [None, None,
      [[Gramext.Stoken ("", "%")],
       Gramext.action
         (fun _ (loc : int * int) -> (loc, Ident.t_mod_int : 'times_div_mod));
       [Gramext.Stoken ("", "/")],
       Gramext.action
         (fun _ (loc : int * int) -> (loc, Ident.t_div : 'times_div_mod));
       [Gramext.Stoken ("", "*")],
       Gramext.action
         (fun _ (loc : int * int) -> (loc, Ident.t_mul : 'times_div_mod))]];
     Grammar.Entry.obj (prog6 : 'prog6 Grammar.Entry.e), None,
     [None, None,
      [[Gramext.Snterm (Grammar.Entry.obj (prog7 : 'prog7 Grammar.Entry.e))],
       Gramext.action (fun (x : 'prog7) (loc : int * int) -> (x : 'prog6));
       [Gramext.Stoken ("LIDENT", "sqrt_float"); Gramext.Sself],
       Gramext.action
         (fun (x : 'prog6) _ (loc : int * int) ->
            (un_op Ident.t_sqrt_float loc x : 'prog6));
       [Gramext.Stoken ("", "-"); Gramext.Sself],
       Gramext.action
         (fun (x : 'prog6) _ (loc : int * int) ->
            (un_op Ident.t_neg loc x : 'prog6))]];
     Grammar.Entry.obj (prog7 : 'prog7 Grammar.Entry.e), None,
     [None, None,
      [[Gramext.Snterm (Grammar.Entry.obj (ident : 'ident Grammar.Entry.e));
        Gramext.Stoken ("", ":");
        Gramext.Snterm
          (Grammar.Entry.obj (program : 'program Grammar.Entry.e))],
       Gramext.action
         (fun (p : 'program) _ (l : 'ident) (loc : int * int) ->
            (without_annot loc
               (Sseq [Slabel (Ident.string l); Sstatement p]) :
             'prog7));
       [Gramext.Stoken ("", "{");
        Gramext.Sopt
          (Gramext.Snterm
             (Grammar.Entry.obj
                (pre_condition : 'pre_condition Grammar.Entry.e)));
        Gramext.Stoken ("", "}");
        Gramext.Snterm
          (Grammar.Entry.obj (program : 'program Grammar.Entry.e));
        Gramext.Stoken ("", "{");
        Gramext.Sopt
          (Gramext.Snterm
             (Grammar.Entry.obj
                (post_condition : 'post_condition Grammar.Entry.e)));
        Gramext.Stoken ("", "}")],
       Gramext.action
         (fun _ (post : 'post_condition option) _ (p : 'program) _
            (pre : 'pre_condition option) _ (loc : int * int) ->
            (mk_prog loc p (list_of_some pre) post : 'prog7));
       [Gramext.Stoken ("", "(");
        Gramext.Snterm
          (Grammar.Entry.obj (program : 'program Grammar.Entry.e));
        Gramext.Slist0
          (Gramext.Snterm (Grammar.Entry.obj (arg : 'arg Grammar.Entry.e)));
        Gramext.Stoken ("", ")")],
       Gramext.action
         (fun _ (args : 'arg list) (p : 'program) _ (loc : int * int) ->
            (if args = [] then p else without_annot loc (app p args) :
             'prog7));
       [Gramext.Stoken ("", "absurd");
        Gramext.Sopt
          (Gramext.Snterm
             (Grammar.Entry.obj (cast : 'cast Grammar.Entry.e)))],
       Gramext.action
         (fun (t : 'cast option) _ (loc : int * int) ->
            (without_annot loc (Sabsurd t) : 'prog7));
       [Gramext.Stoken ("", "try");
        Gramext.Snterm
          (Grammar.Entry.obj (program : 'program Grammar.Entry.e));
        Gramext.Stoken ("", "with");
        Gramext.Slist1sep
          (Gramext.Snterm
             (Grammar.Entry.obj (handler : 'handler Grammar.Entry.e)),
           Gramext.Stoken ("", "|"));
        Gramext.Stoken ("", "end")],
       Gramext.action
         (fun _ (hl : 'handler list) _ (p : 'program) _ (loc : int * int) ->
            (without_annot loc (Stry (p, hl)) : 'prog7));
       [Gramext.Stoken ("", "raise"); Gramext.Stoken ("", "(");
        Gramext.Snterm (Grammar.Entry.obj (ident : 'ident Grammar.Entry.e));
        Gramext.Snterm
          (Grammar.Entry.obj (program : 'program Grammar.Entry.e));
        Gramext.Stoken ("", ")");
        Gramext.Sopt
          (Gramext.Snterm
             (Grammar.Entry.obj (cast : 'cast Grammar.Entry.e)))],
       Gramext.action
         (fun (t : 'cast option) _ (p : 'program) (id : 'ident) _ _
            (loc : int * int) ->
            (without_annot loc (Sraise (id, Some p, t)) : 'prog7));
       [Gramext.Stoken ("", "raise");
        Gramext.Snterm (Grammar.Entry.obj (ident : 'ident Grammar.Entry.e));
        Gramext.Sopt
          (Gramext.Snterm
             (Grammar.Entry.obj (cast : 'cast Grammar.Entry.e)))],
       Gramext.action
         (fun (t : 'cast option) (id : 'ident) _ (loc : int * int) ->
            (without_annot loc (Sraise (id, None, t)) : 'prog7));
       [Gramext.Stoken ("", "let"); Gramext.Stoken ("", "rec");
        Gramext.Snterm (Grammar.Entry.obj (recfun : 'recfun Grammar.Entry.e));
        Gramext.Stoken ("", "in");
        Gramext.Snterm
          (Grammar.Entry.obj (program : 'program Grammar.Entry.e))],
       Gramext.action
         (fun (p2 : 'program) _ (p : 'recfun) _ _ (loc : int * int) ->
            (without_annot loc
               (Sletin (rec_name p, without_annot loc p, p2)) :
             'prog7));
       [Gramext.Stoken ("", "let"); Gramext.Stoken ("", "rec");
        Gramext.Snterm
          (Grammar.Entry.obj (recfun : 'recfun Grammar.Entry.e))],
       Gramext.action
         (fun (p : 'recfun) _ _ (loc : int * int) ->
            (without_annot loc p : 'prog7));
       [Gramext.Stoken ("", "fun");
        Gramext.Snterm
          (Grammar.Entry.obj (binders : 'binders Grammar.Entry.e));
        Gramext.Stoken ("", "->");
        Gramext.Snterm
          (Grammar.Entry.obj (program : 'program Grammar.Entry.e))],
       Gramext.action
         (fun (p : 'program) _ (bl : 'binders) _ (loc : int * int) ->
            (without_annot loc (Slam (bl, p)) : 'prog7));
       [Gramext.Stoken ("", "begin");
        Gramext.Snterm (Grammar.Entry.obj (block : 'block Grammar.Entry.e));
        Gramext.Stoken ("", "end")],
       Gramext.action
         (fun _ (_, b : 'block) _ (loc : int * int) ->
            (without_annot loc (Sseq b) : 'prog7));
       [Gramext.Stoken ("", "let");
        Gramext.Snterm (Grammar.Entry.obj (ident : 'ident Grammar.Entry.e));
        Gramext.Stoken ("", "=");
        Gramext.Snterm
          (Grammar.Entry.obj (program : 'program Grammar.Entry.e));
        Gramext.Stoken ("", "in");
        Gramext.Snterm
          (Grammar.Entry.obj (program : 'program Grammar.Entry.e))],
       Gramext.action
         (fun (p2 : 'program) _ (p1 : 'program) _ (v : 'ident) _
            (loc : int * int) ->
            (without_annot loc (Sletin (v, p1, p2)) : 'prog7));
       [Gramext.Stoken ("", "let");
        Gramext.Snterm (Grammar.Entry.obj (ident : 'ident Grammar.Entry.e));
        Gramext.Stoken ("", "="); Gramext.Stoken ("", "ref");
        Gramext.Snterm
          (Grammar.Entry.obj (program : 'program Grammar.Entry.e));
        Gramext.Stoken ("", "in");
        Gramext.Snterm
          (Grammar.Entry.obj (program : 'program Grammar.Entry.e))],
       Gramext.action
         (fun (p2 : 'program) _ (p1 : 'program) _ _ (v : 'ident) _
            (loc : int * int) ->
            (without_annot loc (Sletref (v, p1, p2)) : 'prog7));
       [Gramext.Stoken ("", "while");
        Gramext.Snterm
          (Grammar.Entry.obj (program : 'program Grammar.Entry.e));
        Gramext.Stoken ("", "do"); Gramext.Stoken ("", "{");
        Gramext.Sopt
          (Gramext.Snterm
             (Grammar.Entry.obj (invariant : 'invariant Grammar.Entry.e)));
        Gramext.Stoken ("LIDENT", "variant");
        Gramext.Snterm
          (Grammar.Entry.obj (variant : 'variant Grammar.Entry.e));
        Gramext.Stoken ("", "}");
        Gramext.Snterm (Grammar.Entry.obj (block : 'block Grammar.Entry.e));
        Gramext.Stoken ("", "done")],
       Gramext.action
         (fun _ (bl_loc, bl : 'block) _ (wf : 'variant) _
            (inv : 'invariant option) _ _ (b : 'program) _
            (loc : int * int) ->
            (without_annot loc
               (Swhile (b, inv, wf, without_annot bl_loc (Sseq bl))) :
             'prog7));
       [Gramext.Stoken ("", "if");
        Gramext.Snterm
          (Grammar.Entry.obj (program : 'program Grammar.Entry.e));
        Gramext.Stoken ("", "then");
        Gramext.Snterm
          (Grammar.Entry.obj (program : 'program Grammar.Entry.e))],
       Gramext.action
         (fun (e2 : 'program) _ (e1 : 'program) _ (loc : int * int) ->
            (without_annot loc
               (Sif (e1, e2, without_annot (no_loc loc) (Sconst ConstUnit))) :
             'prog7));
       [Gramext.Stoken ("", "if");
        Gramext.Snterm
          (Grammar.Entry.obj (program : 'program Grammar.Entry.e));
        Gramext.Stoken ("", "then");
        Gramext.Snterm
          (Grammar.Entry.obj (program : 'program Grammar.Entry.e));
        Gramext.Stoken ("", "else");
        Gramext.Snterm
          (Grammar.Entry.obj (program : 'program Grammar.Entry.e))],
       Gramext.action
         (fun (e3 : 'program) _ (e2 : 'program) _ (e1 : 'program) _
            (loc : int * int) ->
            (without_annot loc (Sif (e1, e2, e3)) : 'prog7));
       [Gramext.Snterm (Grammar.Entry.obj (ident : 'ident Grammar.Entry.e));
        Gramext.Stoken ("", "[");
        Gramext.Snterm
          (Grammar.Entry.obj (program : 'program Grammar.Entry.e));
        Gramext.Stoken ("", "]"); Gramext.Stoken ("", ":=");
        Gramext.Snterm
          (Grammar.Entry.obj (program : 'program Grammar.Entry.e))],
       Gramext.action
         (fun (p : 'program) _ _ (e : 'program) _ (v : 'ident)
            (loc : int * int) ->
            (without_annot loc (Sarrset (true, v, e, p)) : 'prog7));
       [Gramext.Snterm (Grammar.Entry.obj (ident : 'ident Grammar.Entry.e));
        Gramext.Stoken ("", "[");
        Gramext.Snterm
          (Grammar.Entry.obj (program : 'program Grammar.Entry.e));
        Gramext.Stoken ("", "]")],
       Gramext.action
         (fun _ (e : 'program) _ (v : 'ident) (loc : int * int) ->
            (without_annot loc (Sarrget (true, v, e)) : 'prog7));
       [Gramext.Snterm (Grammar.Entry.obj (ident : 'ident Grammar.Entry.e));
        Gramext.Stoken ("", ":=");
        Gramext.Snterm
          (Grammar.Entry.obj (program : 'program Grammar.Entry.e))],
       Gramext.action
         (fun (p : 'program) _ (v : 'ident) (loc : int * int) ->
            (without_annot loc (Srefset (v, p)) : 'prog7));
       [Gramext.Stoken ("", "!");
        Gramext.Snterm (Grammar.Entry.obj (ident : 'ident Grammar.Entry.e))],
       Gramext.action
         (fun (v : 'ident) _ (loc : int * int) ->
            (without_annot loc (Srefget v) : 'prog7));
       [Gramext.Stoken ("", "false")],
       Gramext.action
         (fun _ (loc : int * int) ->
            (without_annot loc (Sconst (ConstBool false)) : 'prog7));
       [Gramext.Stoken ("", "true")],
       Gramext.action
         (fun _ (loc : int * int) ->
            (without_annot loc (Sconst (ConstBool true)) : 'prog7));
       [Gramext.Stoken ("LIDENT", "void")],
       Gramext.action
         (fun _ (loc : int * int) ->
            (without_annot loc (Sconst ConstUnit) : 'prog7));
       [Gramext.Stoken ("FLOAT", "")],
       Gramext.action
         (fun (f : string) (loc : int * int) ->
            (without_annot loc (Sconst (ConstFloat f)) : 'prog7));
       [Gramext.Stoken ("INT", "")],
       Gramext.action
         (fun (n : string) (loc : int * int) ->
            (without_annot loc (Sconst (ConstInt (int_of_string n))) :
             'prog7));
       [Gramext.Snterm (Grammar.Entry.obj (ident : 'ident Grammar.Entry.e))],
       Gramext.action
         (fun (v : 'ident) (loc : int * int) ->
            (without_annot loc (Svar v) : 'prog7))]];
     Grammar.Entry.obj (recfun : 'recfun Grammar.Entry.e), None,
     [None, None,
      [[Gramext.Snterm (Grammar.Entry.obj (ident : 'ident Grammar.Entry.e));
        Gramext.Snterm
          (Grammar.Entry.obj (binders : 'binders Grammar.Entry.e));
        Gramext.Stoken ("", ":");
        Gramext.Snterm (Grammar.Entry.obj (type_v : 'type_v Grammar.Entry.e));
        Gramext.Stoken ("", "{"); Gramext.Stoken ("LIDENT", "variant");
        Gramext.Snterm
          (Grammar.Entry.obj (variant : 'variant Grammar.Entry.e));
        Gramext.Stoken ("", "}"); Gramext.Stoken ("", "=");
        Gramext.Snterm
          (Grammar.Entry.obj (program : 'program Grammar.Entry.e))],
       Gramext.action
         (fun (p : 'program) _ _ (var : 'variant) _ _ (v : 'type_v) _
            (bl : 'binders) (f : 'ident) (loc : int * int) ->
            (Srec (f, bl, v, var, p) : 'recfun))]];
     Grammar.Entry.obj (arg : 'arg Grammar.Entry.e), None,
     [None, None,
      [[Gramext.Snterm
          (Grammar.Entry.obj (program : 'program Grammar.Entry.e))],
       Gramext.action
         (fun (p : 'program) (loc : int * int) -> (Sterm p : 'arg));
       [Gramext.Stoken ("", "'");
        Gramext.Snterm
          (Grammar.Entry.obj (type_v : 'type_v Grammar.Entry.e))],
       Gramext.action
         (fun (t : 'type_v) _ (loc : int * int) -> (Stype t : 'arg))]];
     Grammar.Entry.obj (cast : 'cast Grammar.Entry.e), None,
     [None, None,
      [[Gramext.Stoken ("", ":");
        Gramext.Snterm
          (Grammar.Entry.obj (type_v : 'type_v Grammar.Entry.e))],
       Gramext.action
         (fun (t : 'type_v) _ (loc : int * int) -> (t : 'cast))]];
     Grammar.Entry.obj (block : 'block Grammar.Entry.e), None,
     [None, None,
      [[Gramext.Snterm
          (Grammar.Entry.obj (any_block : 'any_block Grammar.Entry.e))],
       Gramext.action
         (fun (b : 'any_block) (loc : int * int) ->
            (check_block loc b; loc, b : 'block))]];
     Grammar.Entry.obj (any_block : 'any_block Grammar.Entry.e), None,
     [None, None,
      [[Gramext.Snterm
          (Grammar.Entry.obj
             (block_statement : 'block_statement Grammar.Entry.e))],
       Gramext.action
         (fun (s : 'block_statement) (loc : int * int) -> ([s] : 'any_block));
       [Gramext.Snterm
          (Grammar.Entry.obj
             (block_statement : 'block_statement Grammar.Entry.e));
        Gramext.Stoken ("", ";"); Gramext.Sself],
       Gramext.action
         (fun (b : 'any_block) _ (s : 'block_statement) (loc : int * int) ->
            (s :: b : 'any_block))]];
     Grammar.Entry.obj (block_statement : 'block_statement Grammar.Entry.e),
     None,
     [None, None,
      [[Gramext.Snterm
          (Grammar.Entry.obj (program : 'program Grammar.Entry.e))],
       Gramext.action
         (fun (p : 'program) (loc : int * int) ->
            (Sstatement p : 'block_statement));
       [Gramext.Stoken ("LIDENT", "assert"); Gramext.Stoken ("", "{");
        Gramext.Snterm
          (Grammar.Entry.obj (assertion : 'assertion Grammar.Entry.e));
        Gramext.Stoken ("", "}")],
       Gramext.action
         (fun _ (c : 'assertion) _ _ (loc : int * int) ->
            (Sassert c : 'block_statement));
       [Gramext.Stoken ("LIDENT", "label");
        Gramext.Snterm (Grammar.Entry.obj (ident : 'ident Grammar.Entry.e))],
       Gramext.action
         (fun (s : 'ident) _ (loc : int * int) ->
            (Slabel (Ident.string s) : 'block_statement))]];
     Grammar.Entry.obj (handler : 'handler Grammar.Entry.e), None,
     [None, None,
      [[Gramext.Snterm (Grammar.Entry.obj (ident : 'ident Grammar.Entry.e));
        Gramext.Snterm (Grammar.Entry.obj (ident : 'ident Grammar.Entry.e));
        Gramext.Stoken ("", "->");
        Gramext.Snterm
          (Grammar.Entry.obj (program : 'program Grammar.Entry.e))],
       Gramext.action
         (fun (p : 'program) _ (v : 'ident) (x : 'ident) (loc : int * int) ->
            ((x, Some v), p : 'handler));
       [Gramext.Snterm (Grammar.Entry.obj (ident : 'ident Grammar.Entry.e));
        Gramext.Stoken ("", "->");
        Gramext.Snterm
          (Grammar.Entry.obj (program : 'program Grammar.Entry.e))],
       Gramext.action
         (fun (p : 'program) _ (x : 'ident) (loc : int * int) ->
            ((x, None), p : 'handler))]];
     Grammar.Entry.obj (relation : 'relation Grammar.Entry.e), None,
     [None, None,
      [[Gramext.Stoken ("", "<>")],
       Gramext.action
         (fun _ (loc : int * int) -> (loc, Ident.t_neq : 'relation));
       [Gramext.Stoken ("", "=")],
       Gramext.action
         (fun _ (loc : int * int) -> (loc, Ident.t_eq : 'relation));
       [Gramext.Stoken ("", ">=")],
       Gramext.action
         (fun _ (loc : int * int) -> (loc, Ident.t_ge : 'relation));
       [Gramext.Stoken ("", ">")],
       Gramext.action
         (fun _ (loc : int * int) -> (loc, Ident.t_gt : 'relation));
       [Gramext.Stoken ("", "<=")],
       Gramext.action
         (fun _ (loc : int * int) -> (loc, Ident.t_le : 'relation));
       [Gramext.Stoken ("", "<")],
       Gramext.action
         (fun _ (loc : int * int) -> (loc, Ident.t_lt : 'relation))]];
     Grammar.Entry.obj (invariant : 'invariant Grammar.Entry.e), None,
     [None, None,
      [[Gramext.Stoken ("LIDENT", "invariant");
        Gramext.Snterm
          (Grammar.Entry.obj (assertion : 'assertion Grammar.Entry.e))],
       Gramext.action
         (fun (c : 'assertion) _ (loc : int * int) -> (c : 'invariant))]];
     Grammar.Entry.obj (variant : 'variant Grammar.Entry.e), None,
     [None, None,
      [[Gramext.Snterm (Grammar.Entry.obj (lexpr : 'lexpr Grammar.Entry.e))],
       Gramext.action
         (fun (c : 'lexpr) (loc : int * int) ->
            (c, Ident.t_zwf_zero : 'variant));
       [Gramext.Snterm (Grammar.Entry.obj (lexpr : 'lexpr Grammar.Entry.e));
        Gramext.Stoken ("", "for");
        Gramext.Snterm (Grammar.Entry.obj (ident : 'ident Grammar.Entry.e))],
       Gramext.action
         (fun (r : 'ident) _ (c : 'lexpr) (loc : int * int) ->
            (c, r : 'variant))]];
     Grammar.Entry.obj (exception_type : 'exception_type Grammar.Entry.e),
     None,
     [None, None,
      [[Gramext.Stoken ("", "of");
        Gramext.Snterm
          (Grammar.Entry.obj
             (primitive_type : 'primitive_type Grammar.Entry.e))],
       Gramext.action
         (fun (v : 'primitive_type) _ (loc : int * int) ->
            (v : 'exception_type))]];
     Grammar.Entry.obj (decl : 'decl Grammar.Entry.e), None,
     [None, None,
      [[Gramext.Stoken ("", "logic");
        Gramext.Slist1sep
          (Gramext.Snterm
             (Grammar.Entry.obj (ident : 'ident Grammar.Entry.e)),
           Gramext.Stoken ("", ","));
        Gramext.Stoken ("", ":");
        Gramext.Snterm
          (Grammar.Entry.obj (logic_type : 'logic_type Grammar.Entry.e))],
       Gramext.action
         (fun (t : 'logic_type) _ (ids : 'ident list) _ (loc : int * int) ->
            (Logic (loc, ids, t) : 'decl));
       [Gramext.Stoken ("", "exception");
        Gramext.Snterm (Grammar.Entry.obj (ident : 'ident Grammar.Entry.e));
        Gramext.Sopt
          (Gramext.Snterm
             (Grammar.Entry.obj
                (exception_type : 'exception_type Grammar.Entry.e)))],
       Gramext.action
         (fun (v : 'exception_type option) (id : 'ident) _
            (loc : int * int) ->
            (Exception (loc, id, v) : 'decl));
       [Gramext.Stoken ("", "parameter");
        Gramext.Slist1sep
          (Gramext.Snterm
             (Grammar.Entry.obj (ident : 'ident Grammar.Entry.e)),
           Gramext.Stoken ("", ","));
        Gramext.Stoken ("", ":");
        Gramext.Snterm
          (Grammar.Entry.obj (type_v : 'type_v Grammar.Entry.e))],
       Gramext.action
         (fun (v : 'type_v) _ (ids : 'ident list) _ (loc : int * int) ->
            (Parameter (loc, ids, v) : 'decl));
       [Gramext.Stoken ("", "external");
        Gramext.Slist1sep
          (Gramext.Snterm
             (Grammar.Entry.obj (ident : 'ident Grammar.Entry.e)),
           Gramext.Stoken ("", ","));
        Gramext.Stoken ("", ":");
        Gramext.Snterm
          (Grammar.Entry.obj (type_v : 'type_v Grammar.Entry.e))],
       Gramext.action
         (fun (v : 'type_v) _ (ids : 'ident list) _ (loc : int * int) ->
            (External (loc, ids, v) : 'decl));
       [Gramext.Stoken ("", "let"); Gramext.Stoken ("", "rec");
        Gramext.Snterm
          (Grammar.Entry.obj (recfun : 'recfun Grammar.Entry.e))],
       Gramext.action
         (fun (p : 'recfun) _ _ (loc : int * int) ->
            (Program (rec_name p, without_annot loc p) : 'decl));
       [Gramext.Stoken ("", "let");
        Gramext.Snterm (Grammar.Entry.obj (ident : 'ident Grammar.Entry.e));
        Gramext.Snterm
          (Grammar.Entry.obj (binders : 'binders Grammar.Entry.e));
        Gramext.Stoken ("", "=");
        Gramext.Snterm
          (Grammar.Entry.obj (program : 'program Grammar.Entry.e))],
       Gramext.action
         (fun (p : 'program) _ (bl : 'binders) (id : 'ident) _
            (loc : int * int) ->
            (Program (id, without_annot loc (Slam (bl, p))) : 'decl));
       [Gramext.Stoken ("", "let");
        Gramext.Snterm (Grammar.Entry.obj (ident : 'ident Grammar.Entry.e));
        Gramext.Stoken ("", "=");
        Gramext.Snterm
          (Grammar.Entry.obj (program : 'program Grammar.Entry.e))],
       Gramext.action
         (fun (p : 'program) _ (id : 'ident) _ (loc : int * int) ->
            (Program (id, p) : 'decl))]];
     Grammar.Entry.obj (decls : 'decls Grammar.Entry.e), None,
     [None, None,
      [[Gramext.Slist0
          (Gramext.Snterm (Grammar.Entry.obj (decl : 'decl Grammar.Entry.e)));
        Gramext.Stoken ("EOI", "")],
       Gramext.action
         (fun _ (d : 'decl list) (loc : int * int) -> (d : 'decls))]];
     Grammar.Entry.obj (logic_type : 'logic_type Grammar.Entry.e), None,
     [None, None,
      [[Gramext.Slist0sep
          (Gramext.Snterm
             (Grammar.Entry.obj (logic_arg : 'logic_arg Grammar.Entry.e)),
           Gramext.Stoken ("", ","));
        Gramext.Stoken ("", "->");
        Gramext.Snterm
          (Grammar.Entry.obj
             (primitive_type : 'primitive_type Grammar.Entry.e))],
       Gramext.action
         (fun (t : 'primitive_type) _ (b : 'logic_arg list)
            (loc : int * int) ->
            (Function (b, t) : 'logic_type));
       [Gramext.Slist0sep
          (Gramext.Snterm
             (Grammar.Entry.obj (logic_arg : 'logic_arg Grammar.Entry.e)),
           Gramext.Stoken ("", ","));
        Gramext.Stoken ("", "->"); Gramext.Stoken ("LIDENT", "prop")],
       Gramext.action
         (fun _ _ (b : 'logic_arg list) (loc : int * int) ->
            (Predicate b : 'logic_type))]];
     Grammar.Entry.obj (logic_arg : 'logic_arg Grammar.Entry.e), None,
     [None, None,
      [[Gramext.Snterm
          (Grammar.Entry.obj
             (primitive_type : 'primitive_type Grammar.Entry.e));
        Gramext.Stoken ("", "array")],
       Gramext.action
         (fun _ (t : 'primitive_type) (loc : int * int) ->
            (PTarray t : 'logic_arg));
       [Gramext.Snterm
          (Grammar.Entry.obj
             (primitive_type : 'primitive_type Grammar.Entry.e))],
       Gramext.action
         (fun (t : 'primitive_type) (loc : int * int) -> (t : 'logic_arg))]]]

(*s C annotations *)

let c_pre_condition = gec "c_pre_condition"
let c_post_condition = gec "c_post_condition"
let c_spec = gec "c_spec"
let c_loop_annot = gec "c_loop_annot"
let c_pre = gec "c_pre"
let c_variant = gec "c_variant"
let c_post = gec "c_post"
let c_annot = gec "c_annot"

let _ =
  Grammar.extend
    [Grammar.Entry.obj (c_pre_condition : 'c_pre_condition Grammar.Entry.e),
     None,
     [None, None,
      [[Gramext.Stoken ("LIDENT", "pre");
        Gramext.Snterm
          (Grammar.Entry.obj
             (pre_condition : 'pre_condition Grammar.Entry.e))],
       Gramext.action
         (fun (p : 'pre_condition) _ (loc : int * int) ->
            (p : 'c_pre_condition))]];
     Grammar.Entry.obj (c_post_condition : 'c_post_condition Grammar.Entry.e),
     None,
     [None, None,
      [[Gramext.Stoken ("LIDENT", "post");
        Gramext.Snterm
          (Grammar.Entry.obj
             (post_condition : 'post_condition Grammar.Entry.e))],
       Gramext.action
         (fun (q : 'post_condition) _ (loc : int * int) ->
            (q : 'c_post_condition))]];
     Grammar.Entry.obj (c_spec : 'c_spec Grammar.Entry.e), None,
     [None, None,
      [[Gramext.Sopt
          (Gramext.Snterm
             (Grammar.Entry.obj
                (c_pre_condition : 'c_pre_condition Grammar.Entry.e)));
        Gramext.Snterm
          (Grammar.Entry.obj (effects : 'effects Grammar.Entry.e));
        Gramext.Sopt
          (Gramext.Snterm
             (Grammar.Entry.obj
                (c_post_condition : 'c_post_condition Grammar.Entry.e)));
        Gramext.Stoken ("EOI", "")],
       Gramext.action
         (fun _ (q : 'c_post_condition option) (e : 'effects)
            (p : 'c_pre_condition option) (loc : int * int) ->
            (list_of_some p, e, q : 'c_spec))]];
     Grammar.Entry.obj (c_loop_annot : 'c_loop_annot Grammar.Entry.e), None,
     [None, None,
      [[Gramext.Sopt
          (Gramext.Snterm
             (Grammar.Entry.obj (invariant : 'invariant Grammar.Entry.e)));
        Gramext.Stoken ("LIDENT", "variant");
        Gramext.Snterm
          (Grammar.Entry.obj (variant : 'variant Grammar.Entry.e))],
       Gramext.action
         (fun (v : 'variant) _ (i : 'invariant option) (loc : int * int) ->
            (i, v : 'c_loop_annot))]];
     Grammar.Entry.obj (c_pre : 'c_pre Grammar.Entry.e), None,
     [None, None,
      [[Gramext.Sopt
          (Gramext.Snterm
             (Grammar.Entry.obj
                (pre_condition : 'pre_condition Grammar.Entry.e)));
        Gramext.Sopt
          (Gramext.Snterm
             (Grammar.Entry.obj (c_variant : 'c_variant Grammar.Entry.e)))],
       Gramext.action
         (fun (v : 'c_variant option) (p : 'pre_condition option)
            (loc : int * int) ->
            (p, v : 'c_pre))]];
     Grammar.Entry.obj (c_variant : 'c_variant Grammar.Entry.e), None,
     [None, None,
      [[Gramext.Stoken ("LIDENT", "variant");
        Gramext.Snterm
          (Grammar.Entry.obj (variant : 'variant Grammar.Entry.e))],
       Gramext.action
         (fun (v : 'variant) _ (loc : int * int) -> (v : 'c_variant))]];
     Grammar.Entry.obj (c_post : 'c_post Grammar.Entry.e), None,
     [None, None,
      [[Gramext.Sopt
          (Gramext.Snterm
             (Grammar.Entry.obj
                (post_condition : 'post_condition Grammar.Entry.e)))],
       Gramext.action
         (fun (q : 'post_condition option) (loc : int * int) ->
            (q : 'c_post))]];
     Grammar.Entry.obj (c_annot : 'c_annot Grammar.Entry.e), None,
     [None, None,
      [[Gramext.Stoken ("LIDENT", "label");
        Gramext.Snterm (Grammar.Entry.obj (ident : 'ident Grammar.Entry.e))],
       Gramext.action
         (fun (s : 'ident) _ (loc : int * int) ->
            (Slabel (Ident.string s) : 'c_annot));
       [Gramext.Stoken ("LIDENT", "assert");
        Gramext.Snterm
          (Grammar.Entry.obj (assertion : 'assertion Grammar.Entry.e))],
       Gramext.action
         (fun (p : 'assertion) _ (loc : int * int) ->
            (Sassert p : 'c_annot))]]]

type 'a c_parser = int -> string -> 'a

let parse_with_offset f n s =
  let st = Stream.of_string s in with_offset n (Grammar.Entry.parse f) st

let parse_c_spec = parse_with_offset c_spec
let parse_c_pre = parse_with_offset c_pre
let parse_c_post = parse_with_offset c_post
let parse_c_loop_annot = parse_with_offset c_loop_annot
let parse_c_decl = parse_with_offset decl
let parse_c_annot = parse_with_offset c_annot
